"""
Copyright (c) 2022 Beijing Jiaotong University
PhotLab is licensed under [Open Source License].
You can use this software according to the terms and conditions of the [Open Source License].
You may obtain a copy of [Open Source License] at: [https://open.source.license/]

THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.

See the [Open Source License] for more details.

Author: Chunyu Li
Created: 2022/9/6
Supported by: National Key Research and Development Program of China
"""

import numpy as np
from ..optical import linewidth_induced_noise


class PhaseNoise:
    def __init__(self, over_sampling_rate, linewidth, symbol_rate):
        self.over_sampling_rate = over_sampling_rate
        self.linewidth = linewidth
        self.symbol_rate = symbol_rate

    def _add(self, signal):
        phase_noise = linewidth_induced_noise(len(signal), self.over_sampling_rate, self.linewidth,
                                              self.symbol_rate)
        # 添加相位噪声
        return signal * np.exp(1j * phase_noise)

    def add(self, signal_x, signal_y):
        return self._add(signal_x), self._add(signal_y)



"""
Copyright (c) 2022 Beijing Jiaotong University
PhotLab is licensed under [Open Source License].
You can use this software according to the terms and conditions of the [Open Source License].
You may obtain a copy of [Open Source License] at: [https://open.source.license/]

THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.

See the [Open Source License] for more details.

Author: Chunyu Li
Created: 2022/9/6
Supported by: National Key Research and Development Program of China
"""

from ..utils import upsample
from commpy.filters import rrcosfilter
import numpy as np
from scipy.signal import lfilter


class PulseShaper:
    def __init__(self, up_sampling_factor, len_filter, alpha, ts, fs):
        self.up_sampling_factor = up_sampling_factor
        self.len_filter = len_filter
        self.alpha = alpha
        self.ts = ts
        self.fs = fs

        # 产生RRC滤波器
        time_idx, rrc_filter = rrcosfilter(N=128 * self.up_sampling_factor, alpha=self.alpha, Ts=self.ts,
                                           Fs=self.fs)  # up_sampling_factor*128为滤波器的长度
        self.rrc_filter = (rrc_filter * np.sqrt(2))

    def tx_shape(self, signal_x, signal_y):
        # 先进行插0上采样，上采样倍数为 up_sampling_factor
        up_sampled_signal_x = upsample(signal_x, self.up_sampling_factor)
        up_sampled_signal_y = upsample(signal_y, self.up_sampling_factor)

        # 对信号使用RRC滤波器脉冲整形
        rrc_signal_x = lfilter(self.rrc_filter, 1, up_sampled_signal_x, axis=0)
        rrc_signal_y = lfilter(self.rrc_filter, 1, up_sampled_signal_y, axis=0)
        return rrc_signal_x, rrc_signal_y

    def rx_shape(self, signal_x, signal_y):
        signal_x = lfilter(self.rrc_filter, 1, signal_x, axis=0)
        signal_y = lfilter(self.rrc_filter, 1, signal_y, axis=0)
        return signal_x, signal_y



"""
Copyright (c) 2022 Beijing Jiaotong University
PhotLab is licensed under [Open Source License].
You can use this software according to the terms and conditions of the [Open Source License].
You may obtain a copy of [Open Source License] at: [https://open.source.license/]

THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.

See the [Open Source License] for more details.

Author: Chunyu Li
Created: 2022/9/6
Supported by: National Key Research and Development Program of China
"""

from ..optical import bps_hybrid_qam
from ..utils import plot_scatter
import numpy as np


class BPS:
    def __init__(self, num_test_angle, block_size, bits_per_symbol):
        self.num_test_angle = num_test_angle
        self.block_size = block_size
        self.bits_per_symbol = bits_per_symbol

    def restore(self, signal_x, signal_y):
        # BPS算法
        equalization_matrix_x, phase_x = bps_hybrid_qam(np.real(signal_x), np.imag(signal_x), self.num_test_angle,
                                                        self.block_size, self.bits_per_symbol)
        equalization_matrix_y, phase_y = bps_hybrid_qam(np.real(signal_y), np.imag(signal_y), self.num_test_angle,
                                                        self.block_size, self.bits_per_symbol)
        """ normalization 对信号进行归一化，代码过程中存在一个信号的缩放问题，此处将信号恢复回去 """
        if self.bits_per_symbol == 2:
            equalization_matrix_x = equalization_matrix_x * np.sqrt(2)  # 不同调制格式对应的归一化因子不同
            equalization_matrix_y = equalization_matrix_y * np.sqrt(2)
        elif self.bits_per_symbol == 4:
            equalization_matrix_x = equalization_matrix_x * np.sqrt(10)
            equalization_matrix_y = equalization_matrix_y * np.sqrt(10)
        elif self.bits_per_symbol == 5:
            equalization_matrix_x = equalization_matrix_x * np.sqrt(20)
            equalization_matrix_y = equalization_matrix_y * np.sqrt(20)

        plot_scatter(equalization_matrix_x)

        return equalization_matrix_x, equalization_matrix_y



"""
Copyright (c) 2022 Beijing Jiaotong University
PhotLab is licensed under [Open Source License].
You can use this software according to the terms and conditions of the [Open Source License].
You may obtain a copy of [Open Source License] at: [https://open.source.license/]

THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.

See the [Open Source License] for more details.

Author: Chunyu Li
Created: 2022/9/6
Supported by: National Key Research and Development Program of China
"""

import numpy as np
from ..optical import fine_synchronize, ber_count
from commpy.modulation import QAMModem
from phot import logger


class BER:
    def __init__(self, bits_per_symbol):
        self.bits_per_symbol = bits_per_symbol

    def count(self, signal_x, signal_y, prev_signal_x, prev_signal_y):
        """ 计算误码率 """

        """ 再进行一个帧同步，因为经过均衡器会存在符号的一些舍弃，因此在计算误码率（BER）之前需要再一次帧同步 """

        # 对发射端信号跟均衡后信号进行同步
        start_index_x_1 = fine_synchronize(prev_signal_x[:, 0].T, signal_x[0:10000, 0].reshape((1, -1)))

        # 对发射端信号利用自带函数进行移动
        prev_signal_x = np.roll(prev_signal_x, -start_index_x_1)
        prev_signal_y = np.roll(prev_signal_y, -start_index_x_1)

        # 变为16的倍数
        signal_x = signal_x[:-1]
        signal_y = signal_y[:-1]

        # 使得均衡后信号与发射端信号一样长度
        prev_signal_x = prev_signal_x[0:len(signal_x), :]
        prev_signal_y = prev_signal_y[0:len(signal_y), :]

        """ BER COUNT  对信号进行误码率计算，将接收信号与发射信号转化为格雷编码，比较各个码元的正确率 """

        modem = QAMModem(2 ** self.bits_per_symbol)

        tx_bits = modem.demodulate(
            np.concatenate((np.reshape(prev_signal_x, (-1, 1)), np.reshape(prev_signal_y, (-1, 1))), axis=0).ravel(),
            demod_type='hard')
        rx_bits = modem.demodulate(
            np.concatenate((np.reshape(signal_x, (-1, 1)), np.reshape(signal_y, (-1, 1))), axis=0).ravel(),
            demod_type='hard')
        ber, q_db = ber_count(rx_bits, tx_bits)  # 比较码元的正确率
        logger.info('Calculated overall bits error is {:.5f}'.format(ber))



"""
Copyright (c) 2022 Beijing Jiaotong University
PhotLab is licensed under [Open Source License].
You can use this software according to the terms and conditions of the [Open Source License].
You may obtain a copy of [Open Source License] at: [https://open.source.license/]

THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.

See the [Open Source License] for more details.

Author: Chunyu Li
Created: 2022/9/6
Supported by: National Key Research and Development Program of China
"""

from scipy.signal import resample_poly
from ..optical import dac_resolution
import numpy as np


class DacNoise:
    def __init__(self, sampling_rate_awg, sampling_rate, dac_resolution_bits):
        self.sampling_rate_awg = sampling_rate_awg
        self.sampling_rate = sampling_rate
        self.dac_resolution_bits = dac_resolution_bits

    def _add(self, signal):
        # 重采样信号采样率为DAC的采样率，模拟进入DAC
        signal = resample_poly(signal, int(self.sampling_rate_awg), int(self.sampling_rate))

        # 对信号进行量化，模拟加入量化噪声
        signal = dac_resolution(signal, self.dac_resolution_bits)

        # 减去信号量化后的直流，也就是均值
        signal = signal - np.mean(signal)

        # 重采样信号采样率为原来的采样率，模拟出DAC后采样率，resample_poly 为SciPy库的函数
        signal = resample_poly(signal, int(self.sampling_rate), int(self.sampling_rate_awg)).reshape((-1, 1))

        return signal

    def add(self, signal_x, signal_y):
        return self._add(signal_x), self._add(signal_y)



"""
Copyright (c) 2022 Beijing Jiaotong University
PhotLab is licensed under [Open Source License].
You can use this software according to the terms and conditions of the [Open Source License].
You may obtain a copy of [Open Source License] at: [https://open.source.license/]

THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.

See the [Open Source License] for more details.

Author: Chunyu Li
Created: 2022/9/6
Supported by: National Key Research and Development Program of China
"""

import numpy as np


def gram_schmidt_orthogonalize(input_i, input_q):
    # Ensure length of I,Q data is the same!
    length_min = min(len(input_i), len(input_q))  # 调整信号的长度，保证I,Q路的信号长度一致

    input_i = np.reshape(input_i[0:length_min], (length_min, 1))
    input_q = np.reshape(input_q[0:length_min], (length_min, 1))

    # I acted as the reference orientation

    average_power_i = np.mean(input_i ** 2)  # 对I路信号的平方求平均值
    correction_coefficient = np.mean(input_i * input_q)  # 对I,Q两路信号进行叉乘处理，并取平均值
    reference = input_q - (correction_coefficient / average_power_i) * input_i
    average_power_q = np.mean(reference ** 2)
    output_q = reference / np.sqrt(average_power_q)
    output_i = input_i / np.sqrt(average_power_i)

    # 算法GSOP参考文献  Fatadin, Irshaad, Seb J. Savory, and David Ives.
    # "Compensation of quadrature imbalance in an optical QPSK coherent receiver."
    # IEEE Photonics Technology Letters 20.20 (2008): 1733-1735.

    return output_i, output_q



"""
Copyright (c) 2022 Beijing Jiaotong University
PhotLab is licensed under [Open Source License].
You can use this software according to the terms and conditions of the [Open Source License].
You may obtain a copy of [Open Source License] at: [https://open.source.license/]

THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.

See the [Open Source License] for more details.

Author: Chunyu Li
Created: 2022/9/6
Supported by: National Key Research and Development Program of China
"""

import numpy as np


def read_matrix(filepath):
    f = open(filepath, 'r')
    lines = f.readlines()
    arr = np.zeros((len(lines), 1), dtype=complex)
    for idx, line in enumerate(lines):
        line = line.replace('i', 'j')
        arr[idx][0] = complex(line)
    return arr



"""
Copyright (c) 2022 Beijing Jiaotong University
PhotLab is licensed under [Open Source License].
You can use this software according to the terms and conditions of the [Open Source License].
You may obtain a copy of [Open Source License] at: [https://open.source.license/]

THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.

See the [Open Source License] for more details.

Author: Chunyu Li
Created: 2022/9/6
Supported by: National Key Research and Development Program of China
"""

import logging
import sys

# Create a custom logger
logger = logging.getLogger("phot")
logger.setLevel(logging.DEBUG)

# Create handlers
custom_handler = logging.StreamHandler(sys.stdout)


class CustomFormatter(logging.Formatter):
    green = '\033[92m'
    grey = '\33[90m'
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s (%(filename)s:%(lineno)d)"

    FORMATS = {
        logging.DEBUG: green + format + reset,
        logging.INFO: grey + format + reset,
        logging.WARNING: yellow + format + reset,
        logging.ERROR: red + format + reset,
        logging.CRITICAL: bold_red + format + reset
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


# Create formatters and add it to handlers
custom_handler.setFormatter(CustomFormatter())

# Add handlers to the logger
logger.addHandler(custom_handler)



"""
Copyright (c) 2022 Beijing Jiaotong University
PhotLab is licensed under [Open Source License].
You can use this software according to the terms and conditions of the [Open Source License].
You may obtain a copy of [Open Source License] at: [https://open.source.license/]

THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.

See the [Open Source License] for more details.

Author: Chunyu Li
Created: 2022/9/6
Supported by: National Key Research and Development Program of China
"""

import numpy as np


def gen_bits(length, seed=None):
    """
    Generate random 0 1 bits sequence

    Args:
        length: length of generated sequence
        seed: generator seed

    Returns:

    """
    rng = np.random.default_rng(seed=seed)
    data_x = rng.integers(0, 2, (length, 1))  # 采用randint函数随机产生0 1码元序列x
    data_y = rng.integers(0, 2, (length, 1))  # 采用randint函数随机产生0 1码元序列y
    return data_x, data_y



"""
Copyright (c) 2022 Beijing Jiaotong University
PhotLab is licensed under [Open Source License].
You can use this software according to the terms and conditions of the [Open Source License].
You may obtain a copy of [Open Source License] at: [https://open.source.license/]

THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.

See the [Open Source License] for more details.

Author: Chunyu Li
Created: 2022/9/6
Supported by: National Key Research and Development Program of China
"""

import numpy as np
from scipy.signal import resample_poly
from ..optical import adc_resolution, gram_schmidt_orthogonalize, fre_offset_compensation_fft
from phot import logger


def add_freq_offset(signal_x, signal_y, frequency_offset, sampling_rate):
    """ 添加收发端激光器造成的频偏，就是发射端激光器和接收端激光器的中心频率的偏移差 """

    # 2*pi*N*V*T   通过公式计算频偏造成的相位，N表示每个符号对应的序号，[1:length(TxSignal_X)]
    phase_carrier_offset = (
            np.arange(1, len(signal_x) + 1).T * 2 * np.pi * frequency_offset / sampling_rate).reshape((-1, 1))

    # 添加频偏，频偏也可以看作一个相位
    signal_x = signal_x * np.exp(1j * phase_carrier_offset)
    signal_y = signal_y * np.exp(1j * phase_carrier_offset)
    return signal_x, signal_y


def add_iq_imbalance(signal_x, signal_y):
    rx_xi_tem = np.real(signal_x)  # 取信号实部
    rx_xq_tem = np.imag(signal_x)  # 取信号虚部
    rx_yi_tem = np.real(signal_y)  # 取信号实部
    rx_yq_tem = np.imag(signal_y)  # 取信号虚部

    amplitude_imbalance = np.power(10, 3 / 20)  # 10^(3/20)为幅度失衡因子
    phase_imbalance = np.pi * 80 / 180  # 80/180为相位失衡因子

    # 对虚部信号乘一个幅度失衡因子
    rx_xq_tem = rx_xq_tem * amplitude_imbalance
    rx_yq_tem = rx_yq_tem * amplitude_imbalance

    # 对虚部信号添加一个失衡相位,并将实部虚部组合为复数信号
    signal_x = rx_xi_tem + np.exp(1j * phase_imbalance) * rx_xq_tem
    signal_y = rx_yi_tem + np.exp(1j * phase_imbalance) * rx_yq_tem
    return signal_x, signal_y


def add_adc_noise(signal_x, signal_y, sampling_rate, adc_sample_rate, adc_resolution_bits):
    """ 加入ADC的量化噪声 """

    # 重采样改变采样率为ADC采样率，模拟进入ADC
    re_x = resample_poly(signal_x, int(adc_sample_rate), int(sampling_rate))
    re_y = resample_poly(signal_y, int(adc_sample_rate), int(sampling_rate))

    # 对信号量化，添加ADC造成的量化噪声
    re_x = adc_resolution(re_x, adc_resolution_bits)
    re_y = adc_resolution(re_y, adc_resolution_bits)

    # 减去信号量化后的直流
    re_x = re_x - np.mean(re_x)
    re_y = re_y - np.mean(re_y)

    # 将信号采样率重采样为原来的采样率
    re_x = resample_poly(re_x, int(sampling_rate), int(adc_sample_rate))
    re_y = resample_poly(re_y, int(sampling_rate), int(adc_sample_rate))
    return re_x, re_y


def iq_freq_offset_and_compensation(signal_x, signal_y, sampling_rate):
    """ IQ正交化补偿，就是将之前的I/Q失衡的损伤补偿回来 """

    # 利用GSOP算法对I/Q失衡进行补偿，具体算法原理可看函数内部给的参考文献或论文
    rx_xi_tem, rx_xq_tem = gram_schmidt_orthogonalize(np.real(signal_x), np.imag(signal_x))
    rx_yi_tem, rx_yq_tem = gram_schmidt_orthogonalize(np.real(signal_y), np.imag(signal_y))

    # 对补偿后的实部和虚部信号进行重组
    re_x = rx_xi_tem + 1j * rx_xq_tem
    re_y = rx_yi_tem + 1j * rx_yq_tem

    """ 粗糙的频偏估计和补偿，先进行一个频偏的补偿，因为后面有一个帧同步，而帧同步之前需要先对频偏进行补偿，否则帧同步不正确 """

    # 利用FFT-FOE算法对信号的频偏进行估计与补偿
    re_x, re_y, fre_offset = fre_offset_compensation_fft(re_x, re_y, sampling_rate)
    logger.info('Estimated Coarse Frequency offset: {}'.format(fre_offset))
    return re_x, re_y



"""
Copyright (c) 2022 Beijing Jiaotong University
PhotLab is licensed under [Open Source License].
You can use this software according to the terms and conditions of the [Open Source License].
You may obtain a copy of [Open Source License] at: [https://open.source.license/]

THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.

See the [Open Source License] for more details.

Author: Chunyu Li
Created: 2022/9/6
Supported by: National Key Research and Development Program of China
"""

from .logging import logger
from .mathutils import gen_bits
from .plot import plot_scatter, plot_linechart, config
from .dsp import upsample, quantize, volts_to_decibel
from .io import read_matrix
from .gpu import to_gpu, to_cpu, is_gpu



import numpy as np
from numpy.fft import fftshift, fft, ifft


def dbp_compensate(re_x, re_y, power_temp_x, power_temp_y, sampling_rate, span, num_steps, beta2, delta_z, gamma):
    data_length = np.size(re_x)

    omega = np.arange(-sampling_rate / 2, sampling_rate / 2, sampling_rate / data_length)
    omega = np.reshape(omega, (-1, 1))

    # 信号功率
    power_temp2_x = np.mean(np.abs(re_x) ** 2)
    power_temp2_y = np.mean(np.abs(re_y) ** 2)

    re_x = re_x / (np.sqrt(power_temp2_x / power_temp_x))
    re_y = re_y / (np.sqrt(power_temp2_y / power_temp_y))

    data_fft_x = fftshift(fft(re_x, axis=0), axes=0) / np.size(re_x)  # 将数据变到频率并将零值移到原点
    data_fft_y = fftshift(fft(re_y, axis=0), axes=0) / np.size(re_y)

    for i in range(span):
        for j in range(num_steps):
            # 1/2 光纤色散
            # X 偏振
            data_fft_x = data_fft_x * np.exp(1j * (beta2 / 2) * (omega ** 2) * (delta_z / 2))
            # Y 偏振
            data_fft_y = data_fft_y * np.exp(1j * (beta2 / 2) * (omega ** 2) * (delta_z / 2))

            # 光纤非线性
            # X 偏振
            data_t_x = ifft(fftshift(data_fft_x, axes=0), axis=0) * np.size(data_fft_x)
            # Y 偏振
            data_t_y = ifft(fftshift(data_fft_y, axes=0), axis=0) * np.size(data_fft_y)
            # X 偏振
            data_t_x = data_t_x * np.exp(-8 / 9 * 1j * gamma * ((abs(data_t_y) ** 2) + (abs(data_t_x) ** 2)) * delta_z)
            # Y 偏振
            data_t_y = data_t_y * np.exp(-8 / 9 * 1j * gamma * ((abs(data_t_y) ** 2) + (abs(data_t_x) ** 2)) * delta_z)

            data_fft_x = fftshift(fft(data_t_x, axis=0), axes=0) / np.size(data_t_x)
            data_fft_y = fftshift(fft(data_t_y, axis=0), axes=0) / np.size(data_t_y)

            # 1/2 光纤色散
            # X 偏振
            data_fft_x = data_fft_x * np.exp(1j * (beta2 / 2) * (omega ** 2) * (delta_z / 2))
            # Y 偏振
            data_fft_y = data_fft_y * np.exp(1j * (beta2 / 2) * (omega ** 2) * (delta_z / 2))

    re_x = ifft(fftshift(data_fft_x, axes=0), axis=0) * np.size(data_fft_x)
    re_y = ifft(fftshift(data_fft_y, axes=0), axis=0) * np.size(data_fft_y)

    return re_x, re_y



"""
Copyright (c) 2022 Beijing Jiaotong University
PhotLab is licensed under [Open Source License].
You can use this software according to the terms and conditions of the [Open Source License].
You may obtain a copy of [Open Source License] at: [https://open.source.license/]

THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.

See the [Open Source License] for more details.

Author: Chunyu Li
Created: 2022/9/6
Supported by: National Key Research and Development Program of China
"""

import numpy as np
import matplotlib.pyplot as plt

_plot = True


def config(plot=True):
    global _plot
    _plot = plot


def plot_scatter(sig: np.ndarray, pt_size=5) -> None:
    if not _plot:
        return
    axis_x = np.real(sig).ravel()
    axis_y = np.imag(sig).ravel()
    # plt.figure(figsize=(10, 10), layout='constrained')
    plt.figure(figsize=(10, 10))
    plt.scatter(axis_y, axis_x, s=pt_size)
    plt.xlabel('Real')
    plt.ylabel('Imag')
    plt.title('Scatter Plot')
    plt.grid()
    plt.show()


def plot_linechart(x, y, xlabel=None, ylabel=None, title=None):
    if not _plot:
        return
    # plt.figure(figsize=(10, 10), layout='constrained')
    plt.figure(figsize=(10, 10))
    plt.plot(x, y, color=None)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.title(title)
    plt.show()



"""
Copyright (c) 2022 Beijing Jiaotong University
PhotLab is licensed under [Open Source License].
You can use this software according to the terms and conditions of the [Open Source License].
You may obtain a copy of [Open Source License] at: [https://open.source.license/]

THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.

See the [Open Source License] for more details.

Author: Chunyu Li
Created: 2022/9/6
Supported by: National Key Research and Development Program of China
"""

import numpy as np
from numpy.fft import fft, fftshift
from phot.utils import plot_linechart, volts_to_decibel


def fre_offset_compensation_fft(input_complex_signal_x, input_complex_signal_y, sample_rate):
    """

    Args:
        input_complex_signal_x: input complex samples at 1sample/symbol for FFT at x polarization
        input_complex_signal_y: input complex samples at 1sample/symbol for FFT at y polarization
        sample_rate:

    Returns:

    """

    # 算法 FFT-FOE 参考文献  Selmi, Mehrez, Yves Jaouen, and Philippe Ciblat.
    # "Accurate digital frequency offset estimator for coherent PolMux QAM transmission systems."
    # 2009 35th European Conference on Optical Communication. IEEE, 2009.

    # frequency offset estimation
    fft_size = len(input_complex_signal_x)  # 调整信号行列，将信号转为1列

    # 将信号转为1列
    input_complex_signal_x = np.reshape(input_complex_signal_x, (-1, 1))
    input_complex_signal_y = np.reshape(input_complex_signal_y, (-1, 1))

    # 取信号的4次方
    complex_signal_4th_power_x = input_complex_signal_x ** 4
    complex_signal_4th_power_y = input_complex_signal_y ** 4

    # 取信号的傅里叶变换后，求功率
    power_spectrum_x = fftshift(np.square(np.abs(fft(complex_signal_4th_power_x, axis=0))))
    power_spectrum_y = fftshift(np.square(np.abs(fft(complex_signal_4th_power_y, axis=0))))

    # PowerSpectrum_x=fftshift((abs(fft(ComplexSignal_4thPower_x)).^2).'); %% 取信号的傅里叶变换后，求功率

    power_spectrum = power_spectrum_x + power_spectrum_y  # 对X/Y偏振的功率谱相加

    # step为float型时，arange函数会产生精度问题，故转而采用linspace
    # frequency_scale = np.arange(-sample_rate / 8, sample_rate / 8, sample_rate / fft_size / 4).reshape(
    #     (-1, 1))  # Hz  列出信号的频率坐标

    # Hz  列出信号的频率坐标
    frequency_scale = np.linspace(-sample_rate / 8, sample_rate / 8, fft_size + 1)
    frequency_scale = np.delete(frequency_scale, -1).reshape((-1, 1))

    plot_linechart(frequency_scale * 1e-9, volts_to_decibel(power_spectrum))  # 画图

    max_index = np.argmax(power_spectrum)  # 找到最大的功率谱对应的位置
    fre_offset = frequency_scale[max_index][0]  # 找到最大的功率谱对应的频率坐标就是频偏
    phase_fre_offset = 2 * np.pi * fre_offset / sample_rate  # rad  利用公式计算频偏带来的相位
    foc_vector = np.reshape(np.conj(np.arange(1, len(input_complex_signal_x) + 1)).T * phase_fre_offset,
                            (-1, 1))  # 对信号添加频偏带来的相位

    output_x = input_complex_signal_x * np.exp(-1j * foc_vector)
    output_y = input_complex_signal_y * np.exp(-1j * foc_vector)

    return output_x, output_y, fre_offset



"""
Copyright (c) 2022 Beijing Jiaotong University
PhotLab is licensed under [Open Source License].
You can use this software according to the terms and conditions of the [Open Source License].
You may obtain a copy of [Open Source License] at: [https://open.source.license/]

THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.

See the [Open Source License] for more details.

Author: Chunyu Li
Created: 2022/9/6
Supported by: National Key Research and Development Program of China
"""

import numpy as np
from ..optical import cma_rde, fre_offset_compensation_fft
from ..utils import plot_scatter
from phot import logger


class AdaptiveEqualizer:
    def __init__(self, num_tap, cma_convergence, ref_power_cma, step_size_cma, step_size_rde, up_sampling_factor,
                 bits_per_symbol, total_baud):
        self.num_tap = num_tap
        self.cma_convergence = cma_convergence
        self.ref_power_cma = ref_power_cma
        self.step_size_cma = step_size_cma
        self.step_size_rde = step_size_rde
        self.up_sampling_factor = up_sampling_factor
        self.bits_per_symbol = bits_per_symbol
        self.total_baud = total_baud

    def equalize(self, signal_x, signal_y):
        input_x_i = np.real(signal_x)  # 求出接收端X偏振的实部信号
        input_x_q = np.imag(signal_x)  # 求出接收端X偏振的虚部信号
        input_y_i = np.real(signal_y)  # 求出接收端Y偏振的实部信号
        input_y_q = np.imag(signal_y)  # 求出接收端Y偏振的实部信号

        # 对信号采用CMA-RDE进行自适应均衡
        equalization_matrix_x, equalization_matrix_y = cma_rde(input_x_i, input_x_q, input_y_i, input_y_q, self.num_tap,
                                                               self.cma_convergence, self.ref_power_cma,
                                                               self.step_size_cma, self.step_size_rde,
                                                               self.up_sampling_factor, self.bits_per_symbol)
        plot_scatter(equalization_matrix_x, pt_size=1)

        # 此处均衡器内部存在一个下采样，因此均衡器出来后信号回到一个符号一个样本的采样率，也就是现在的采样率等于符号率

        """ 均衡后进行精确的频偏估计和补偿 采用FFT-FOE算法，与前面的粗估计一样，防止前面粗估计没补偿完全，此处做一个补充 """

        # 利用FFT-FOE算法对信号的频偏进行估计与补偿
        equalization_matrix_x, equalization_matrix_y, fre_offset = fre_offset_compensation_fft(equalization_matrix_x,
                                                                                               equalization_matrix_y,
                                                                                               self.total_baud)
        logger.info('Estimated Accurate Frequency offset: {}'.format(fre_offset))

        return equalization_matrix_x, equalization_matrix_y



"""
Copyright (c) 2022 Beijing Jiaotong University
PhotLab is licensed under [Open Source License].
You can use this software according to the terms and conditions of the [Open Source License].
You may obtain a copy of [Open Source License] at: [https://open.source.license/]

THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.

See the [Open Source License] for more details.

Author: Chunyu Li
Created: 2022/9/6
Supported by: National Key Research and Development Program of China
"""

import numpy as np
from numpy import sqrt
from math import floor, ceil
from numba import njit, prange


def bps_hybrid_qam(input_i, input_q, num_test_angle, block_size, bits_per_symbol):
    """
    此处显示有关此函数的摘要
    采用BPS算法，具体参考文献为：
    Pfau, Timo, Sebastian Hoffmann, and Reinhold Noé.
    "Hardware-efficient coherent digital receiver concept with feedforward carrier recovery for  M-QAM constellations."
    Journal of Lightwave Technology 27.8 (2009): 989-999.
    Args:
        input_i:
        input_q:
        num_test_angle:
        block_size:
        bits_per_symbol:

    Returns:

    """
    # 调整信号矩阵为1列
    input_i = np.reshape(input_i, (-1, 1))
    input_q = np.reshape(input_q, (-1, 1))

    norm_factor = np.sqrt(np.mean(np.abs(input_i) ** 2 + np.abs(input_q) ** 2))  # 进行归一化
    input_i = input_i / norm_factor
    input_q = input_q / norm_factor

    complex_signal = input_i + 1j * input_q  # 组合为复信号
    total_number = len(complex_signal)  # 计算信号长度

    # 得到了一个用测试角旋转后的矩阵
    signal_with_angle = _angle_matrix(total_number, num_test_angle, complex_signal)

    constellation_hy = _init_constellation(bits_per_symbol)

    # 判决
    signal_decided = np.zeros((total_number, num_test_angle), dtype=complex)
    for poinm in range(num_test_angle):
        signal_decided[:, poinm] = decision_hybrid_qam(np.reshape(signal_with_angle[:, poinm], (1, -1)),
                                                       constellation_hy)  # 对预补偿后的信号进行判决

    # 得到距离矩阵
    distance_matrix = signal_with_angle - signal_decided
    distance_matrix_square = np.abs(distance_matrix) ** 2  # 求得判决后的信号与预补偿信号的差值功率

    error = _error_matrix(total_number, num_test_angle, block_size, distance_matrix_square)

    min_result = _calculate_min_result(total_number, error)

    phase = np.zeros((total_number, 1))
    for i in range(total_number):
        phase[i][0] = (min_result[i][1] / num_test_angle - 0.5) * np.pi / 2  # 得到测试角序号对应的测试角

    phase_est_uwapx_compress = np.zeros((len(phase), 1))
    c = np.zeros((len(phase), 1))

    # phase unwrap 用来判断相位是不是属于其它象限，如果是的话将偏移角度变化，对估计的相位进行展开，解决周期滑动
    for i in range(1, len(phase)):
        if abs(phase[i][0] - phase[i - 1][0]) > np.pi / 4:  # 判断前后相位是否相差大于pi/4，大于的话需要加pi/2
            c[i][0] = c[i - 1][0] - np.sign(phase[i][0] - phase[i - 1][0]) * np.pi / 2
        else:
            c[i][0] = c[i - 1][0]
        phase_est_uwapx_compress[i][0] = phase[i][0] + c[i][0]  # 将前后相位进行展开

    output_signal = complex_signal * np.exp(1j * phase_est_uwapx_compress)  # 对信号进行相位噪声补偿
    return output_signal, phase_est_uwapx_compress


@njit
def _calculate_min_result(total_number, error):
    # 通过最小距离来得到估计得相位值
    # decision_fist_stage = np.zeros((total_number, 1), dtype=np.complex128)
    # signal_first_stage = np.zeros((total_number, 1), dtype=np.complex128)
    min_result = np.zeros((total_number, 2))

    for inr in range(total_number):
        # 后面error每一行所得的最小值的列号，就是第几个测试角是正确估计的相位噪声所对应的测试角的序号
        min_result[inr][0] = np.min(error[inr, :])
        min_result[inr][1] = np.argmin(error[inr, :])
        # decision_fist_stage[inr][0] = signal_decided[inr][int(min_result[inr][1])]  # 这是理想的点
        # signal_first_stage[inr][0] = signal_with_angle[inr][int(min_result[inr][1])]  # 这是对应的角度旋转后的符号
    return min_result


@njit(parallel=True)
def _angle_matrix(total_number, num_test_angle, complex_signal):
    # 第一步 测试角
    signal_with_angle = np.zeros((total_number, num_test_angle), dtype=np.complex128)
    for ind in prange(total_number):
        for point in prange(num_test_angle):
            # [-0.5pi,0.5pi]:由于QAM信号的PI/2的旋转对称性， 使用不同测试角对信号进行预补偿
            signal_with_angle[ind][point] = complex_signal[ind][0] * np.exp(
                1j * (point / num_test_angle - 0.5) * np.pi / 2)
    # 得到了一个用测试角旋转后的矩阵

    return signal_with_angle


def _init_constellation(bits_per_symbol):
    # 列出各个调制格式下的原始星座点
    constellation_4qam = np.array([-1 - 1j, -1 + 1j, 1 - 1j, 1 + 1j])

    constellation_8qam = np.array([-1 - 1j, -1 + 1j, 1 - 1j, 1 + 1j, -1 - sqrt(3), 1 + sqrt(3), (-1 - sqrt(3)) * 1j,
                                   (1 + sqrt(3)) * 1j])

    constellation_16qam = np.array([-3 - 3j, -3 - 1j, -3 + 1j, -3 + 3j, -1 - 3j, -1 - 1j, -1 + 1j, -1 + 3j,
                                    1 - 3j, 1 - 1j, 1 + 1j, 1 + 3j, 3 - 3j, 3 - 1j, 3 + 1j, 3 + 3j])

    constellation_32qam = np.array(
        [5 + 3j, 5 + 1j, 5 - 1j, 5 - 3j, 3 + 5j, 3 + 3j, 3 + 1j, 3 - 1j, 3 - 3j, 3 - 5j, 1 + 5j,
         1 + 3j, 1 + 1j, 1 - 1j, 1 - 3j, 1 - 5j, -1 + 5j, -1 + 3j, -1 + 1j, -1 - 1j, -1 - 3j, -1 - 5j,
         -3 + 5j, -3 + 3j, -3 + 1j, -3 - 1j, -3 - 3j, -3 - 5j, -5 + 3j, -5 + 1j, -5 - 1j, -5 - 3j])

    constellation_64qam = np.array(
        [7 + 7j, 7 + 5j, 7 + 3j, 7 + 1j, 7 - 1j, 7 - 3j, 7 - 5j, 7 - 7j, 5 + 7j, 5 + 5j, 5 + 3j, 5 + 1j, 5 - 1j, 5 - 3j,
         5 - 5j, 5 - 7j, 3 + 7j, 3 + 5j, 3 + 3j, 3 + 1j, 3 - 1j, 3 - 3j, 3 - 5j, 3 - 7j, 1 + 7j, 1 + 5j, 1 + 3j, 1 + 1j,
         1 - 1j, 1 - 3j, 1 - 5j, 1 - 7j, -1 + 7j, -1 + 5j, -1 + 3j, -1 + 1j, -1 - 1j, -1 - 3j, -1 - 5j, -1 - 7j,
         -3 + 7j, -3 + 5j, -3 + 3j, -3 + 1j, -3 - 1j, -3 - 3j, -3 - 5j, -3 - 7j, -5 + 7j, -5 + 5j, -5 + 3j, -5 + 1j,
         -5 - 1j, -5 - 3j, -5 - 5j, -5 - 7j, -7 + 7j, -7 + 5j, -7 + 3j, -7 + 1j, -7 - 1j, -7 - 3j, -7 - 5j, -7 - 7j])

    # 对星座点进行归一化，因为接收信号归一化了，这里进行对应
    if bits_per_symbol == 2:
        constellation_hy = constellation_4qam / sqrt(2)
    elif bits_per_symbol == 3:
        constellation_hy = constellation_8qam / sqrt(3 + sqrt(3))
    elif bits_per_symbol == 4:
        constellation_hy = constellation_16qam / sqrt(10)
    elif bits_per_symbol == 5:
        constellation_hy = constellation_32qam / sqrt(20)
    elif bits_per_symbol == 6:
        constellation_hy = constellation_64qam / sqrt(42)
    else:
        raise RuntimeError('Unknown bits number')

    return constellation_hy


@njit(parallel=True)
def _error_matrix(total_number, num_test_angle, block_size, distance_matrix_square):
    error = np.zeros((total_number, num_test_angle))  # 得到误差矩

    for poinn in prange(num_test_angle):
        for inn in prange(total_number):
            block_start = inn - ceil(block_size / 2) + 1  # 计算滑动窗口左边界
            block_end = inn + floor(block_size / 2)  # 计算滑动窗口右边界
            if block_start < 0:
                block_start = 0
            if block_end > total_number:
                block_end = total_number
            error[inn][poinn] = np.sum(distance_matrix_square[block_start:block_end, poinn])  # 计算不同测试角下，块长内误差和
    return error


@njit(parallel=True)
def decision_hybrid_qam(input_signal, constellation_hy):
    # 判断每个信号与哪个星座点的欧式距离最小，则判决这个信号为距离最小的星座点

    m, n = input_signal.shape

    # if not utils.is_gpu(constellation_hy):
    #     constellation_hy = utils.to_gpu(constellation_hy)

    # input_signal_temp = np.ravel(input_signal, order='F')
    # with Pool(processes=8) as p:
    #     out = p.map(_compute_distance, input_signal_temp)
    # out = np.reshape(out, (m, n), order='F')

    out = np.zeros((m, n), dtype=np.complex128)
    for i in prange(m):
        for j in prange(n):
            min_index = np.argmin(np.abs(constellation_hy - input_signal[i][j]))
            out[i][j] = constellation_hy[min_index]

    return out



"""
Copyright (c) 2022 Beijing Jiaotong University
PhotLab is licensed under [Open Source License].
You can use this software according to the terms and conditions of the [Open Source License].
You may obtain a copy of [Open Source License] at: [https://open.source.license/]

THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.

See the [Open Source License] for more details.

Author: Chunyu Li
Created: 2022/9/6
Supported by: National Key Research and Development Program of China
"""

import numpy as np
from phot.utils import quantize


def dac_resolution(signal, resolution_bit):
    """
    DAC Resolution
    Args:
        signal: 输入信号
        resolution_bit: 量化等级

    Returns:
        out: 输出信号
    """
    signal_i = np.real(signal)
    signal_q = np.imag(signal)

    # 把信号分为了实部虚部，分别计算实部虚部最大的长度（定义量化区间）
    max_i = np.max(np.abs(signal_i))
    max_q = np.max(np.abs(signal_q))
    a = min(max_i, max_q)  # 取两个值的最小值为A

    # 每个区间映射的离散数值计算量化值，A的值为1.44 区间[-1.44,1.44]一共分为2^resolution_bit 份
    codebook = np.linspace(-a, a, 2 ** resolution_bit)

    # 分割向量，计算对量化值的分割等级   取中间值 中间值的数量比总体少一
    partition = codebook - (codebook[1] - codebook[0]) / 2

    partition = np.delete(partition, 0)  # 256变为255

    # 量化过后的输出，实部虚部均进行量化
    output_i, samples_quant_i = quantize(signal_i.ravel(), partition, codebook)
    output_q, samples_quant_q = quantize(signal_q.ravel(), partition, codebook)

    out = output_i + 1j * output_q  # 输出信号=实部+j虚部

    # [index等级索引,quants按照索引取到的量化值codebook] =quantiz(sig,partit ion 分割等级 ,codebook量化值)
    # index sig 中数值落入到partition的某个区间对应的编号，区间编号从0开始
    # quants 是 sig 中数值对应的量化后的离散数值，所有离散取值的集合是 codebook 。
    # 因为 index 是从0开始编号，所以，quants 与 index 的具体对应关系是：quants(k) = codebook(index(k) + 1)

    return out


def adc_resolution(signal, resolution_bit) -> np.ndarray:
    """
    ADC Resolution
    Args:
        signal: 输入信号
        resolution_bit: 量化等级

    Returns:
        out: 输出信号
    """
    signal_i = np.real(signal)
    signal_q = np.imag(signal)

    # 把信号分为了实部虚部，分别计算实部虚部最大的长度（定义量化区间）
    max_i = np.max(np.abs(signal_i))
    max_q = np.max(np.abs(signal_q))
    a = min(max_i, max_q)  # 取两个值的最小值为A

    # 每个区间映射的离散数值计算量化值，A的值为1.44 区间[-1.44,1.44]一共分为2^resolution_bit 份
    codebook = np.linspace(-a, a, 2 ** resolution_bit)

    # 分割向量，计算对量化值的分割等级   取中间值 中间值的数量比总体少一
    partition = codebook - (codebook[1] - codebook[0]) / 2

    partition = np.delete(partition, 0)  # 256变为255

    # 量化过后的输出，实部虚部均进行量化
    output_i, samples_quant_i = quantize(signal_i.ravel(), partition, codebook)
    output_q, samples_quant_q = quantize(signal_q.ravel(), partition, codebook)

    out = output_i + 1j * output_q  # 输出信号=实部+j虚部

    # [index等级索引,quants按照索引取到的量化值codebook] =quantiz(sig,partit ion 分割等级 ,codebook量化值)
    # index sig 中数值落入到partition的某个区间对应的编号，区间编号从0开始
    # quants 是 sig 中数值对应的量化后的离散数值，所有离散取值的集合是 codebook 。
    # 因为 index 是从0开始编号，所以，quants 与 index 的具体对应关系是：quants(k) = codebook(index(k) + 1)

    return np.reshape(out, (-1, 1))



"""
Copyright (c) 2022 Beijing Jiaotong University
PhotLab is licensed under [Open Source License].
You can use this software according to the terms and conditions of the [Open Source License].
You may obtain a copy of [Open Source License] at: [https://open.source.license/]

THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.

See the [Open Source License] for more details.

Author: Chunyu Li
Created: 2022/9/6
Supported by: National Key Research and Development Program of China
"""

from ..optical import fine_synchronize
import numpy as np
from phot import logger


class FrameSyncer:
    def __init__(self, up_sampling_factor):
        self.up_sampling_factor = up_sampling_factor

    def sync(self, signal_x, signal_y, prev_signal_x, prev_signal_y):
        """ 帧同步，寻找与发射端原始信号头部对应的符号 """

        # 对信号进行帧同步，找出接收信号与发射信号对准的开头
        start_index_x_1 = fine_synchronize(signal_x[0:10000 * self.up_sampling_factor:self.up_sampling_factor].T,
                                           prev_signal_x[0:2000].T)
        start_index_y_1 = fine_synchronize(signal_y[0:10000 * self.up_sampling_factor:self.up_sampling_factor].T,
                                           prev_signal_y[0:2000].T)

        logger.info('两个偏振第一次对准的帧头')
        logger.info('Start_Index_X_1: {} Start_Index_Y_1: {}'.format(start_index_x_1, start_index_y_1))

        # 将帧头位置前的信号去除，以便发射端信号与接收端信号的头部对准
        signal_x = np.delete(signal_x, np.arange(0, start_index_x_1 * self.up_sampling_factor))
        signal_y = np.delete(signal_y, np.arange(0, start_index_y_1 * self.up_sampling_factor))

        # 去掉接收信号的尾部部分信号
        signal_x = signal_x[:-1000]
        signal_y = signal_y[:-1000]

        # 调整发射端信号与接收端信号的长度
        prev_signal_x = prev_signal_x[0:int(np.floor(len(signal_x) / self.up_sampling_factor))]
        prev_signal_y = prev_signal_y[0:int(np.floor(len(signal_y) / self.up_sampling_factor))]

        return signal_x, signal_y, prev_signal_x, prev_signal_y



"""
Copyright (c) 2022 Beijing Jiaotong University
PhotLab is licensed under [Open Source License].
You can use this software according to the terms and conditions of the [Open Source License].
You may obtain a copy of [Open Source License] at: [https://open.source.license/]

THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.

See the [Open Source License] for more details.

Author: Chunyu Li
Created: 2022/9/6
Supported by: National Key Research and Development Program of China
"""

from commpy.modulation import QAMModem


class Modem:
    def __init__(self, bits_per_symbol):
        self.bits_per_symbol = bits_per_symbol
        pass

    def modulate(self, data_x, data_y):
        # QAM调制器
        modem = QAMModem(2 ** self.bits_per_symbol)
        symbols_x = modem.modulate(data_x).reshape((-1, 1))  # X偏振信号
        symbols_y = modem.modulate(data_y).reshape((-1, 1))  # Y偏振信号
        return symbols_x, symbols_y



"""
Copyright (c) 2022 Beijing Jiaotong University
PhotLab is licensed under [Open Source License].
You can use this software according to the terms and conditions of the [Open Source License].
You may obtain a copy of [Open Source License] at: [https://open.source.license/]

THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.

See the [Open Source License] for more details.

Author: Chunyu Li
Created: 2022/9/6
Supported by: National Key Research and Development Program of China
"""

from .logging import logger
import numpy as np


def to_gpu(array):
    try:
        import cupy as cp
        if isinstance(array, cp._core.core.ndarray):
            return array
        return cp.asarray(array)
    except ModuleNotFoundError:
        logger.warning("CUDA is not available, use CPU computation instead")
        return array


def to_cpu(array):
    if isinstance(array, np.ndarray):
        return array
    try:
        import cupy as cp
        return cp.asnumpy(array)
    except ModuleNotFoundError:
        logger.warning("CUDA is not available, use CPU computation instead")
        return array


def is_gpu(array):
    try:
        import cupy as cp
        if isinstance(array, cp._core.core.ndarray):
            return True
        else:
            return False
    except ModuleNotFoundError:
        logger.warning("CUDA is not available, use CPU computation instead")
        return array



"""
Copyright (c) 2022 Beijing Jiaotong University
PhotLab is licensed under [Open Source License].
You can use this software according to the terms and conditions of the [Open Source License].
You may obtain a copy of [Open Source License] at: [https://open.source.license/]

THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.

See the [Open Source License] for more details.

Author: Chunyu Li
Created: 2022/9/6
Supported by: National Key Research and Development Program of China
"""

import numpy as np
from numba import jit


# https://stackoverflow.com/questions/15983986/convert-quantiz-function-to-python
@jit(nopython=True)
def quantize(signal, partitions, codebook):
    indices = np.zeros(len(signal))
    quanta = np.zeros(len(signal))
    for i, datum in enumerate(signal):
        index = 0
        while index < len(partitions) and datum > partitions[index]:
            index += 1
        indices[i] = index
        quanta[i] = codebook[index]
    return indices, quanta


def upsample(array, n):
    if isinstance(n, float):
        if n.is_integer():
            n = int(n)
        else:
            raise RuntimeError('Up-sample coefficient must be integer')
    rows, cols = array.shape
    result = np.zeros((rows * n, cols), dtype=complex)
    for i in range(rows):
        for j in range(cols):
            result[i * n][j] = array[i][j]
    return result


def volts_to_decibel(x):
    return 20 * np.log10(x)



"""
Copyright (c) 2022 Beijing Jiaotong University
PhotLab is licensed under [Open Source License].
You can use this software according to the terms and conditions of the [Open Source License].
You may obtain a copy of [Open Source License] at: [https://open.source.license/]

THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.

See the [Open Source License] for more details.

Author: Chunyu Li
Created: 2022/9/6
Supported by: National Key Research and Development Program of China
"""

import numpy as np
from math import floor, ceil


def cma_rde(input_x_i, input_x_q, input_y_i, input_y_q, num_tap, num_update_cma, ref_power_cma, step_size_cma,
            step_size_rde, samples_per_symbol, bits_per_symbol):
    """
    Function description:  Common CMA with arbitrary tap number and sampling rate
    CMA算法原理可参考：Faruk, Md Saifuddin, and Seb J. Savory. "Digital signal processing for coherent transceivers employing multilevel formats."
    Journal of Lightwave Technology 35.5 (2017): 1125-1141.
    Args:
        input_x_i:
        input_x_q:
        input_y_i:
        input_y_q:
        num_tap: number of FIR taps for CMA, must be an odd number
        num_update_cma: Iteration count for CMA
        ref_power_cma: the modulus^2 for CMA
        step_size_cma: step size for CMA
        step_size_rde:
        samples_per_symbol: oversampling rate
        bits_per_symbol:

    Returns:
        output_x
        output_y
    """
    # Initialize the tap coefficients  对均衡器的抽头系数进行初始化，设置中间的系数为1，其他为0
    initial_weight_xx = np.zeros((num_tap, 1))
    initial_weight_yx = np.zeros((num_tap, 1))
    initial_weight_xy = np.zeros((num_tap, 1))
    initial_weight_yy = np.zeros((num_tap, 1))
    initial_weight_xx[int(floor(num_tap - 1) / 2)] = 1
    initial_weight_yy[int(floor(num_tap - 1) / 2)] = 1

    # Normalization
    norm_factor = np.sqrt(
        2 / np.mean(input_x_i ** 2 + input_x_q ** 2 + input_y_i ** 2 + input_y_q ** 2))  # 对信号进行归一化，加快运算速度
    input_x_i = input_x_i * norm_factor
    input_x_q = input_x_q * norm_factor
    input_y_i = input_y_i * norm_factor
    input_y_q = input_y_q * norm_factor

    # I&Q combined to complex signals
    qpsk_x = input_x_i + 1j * input_x_q  # 对实部虚部进行组合为复数信号
    qpsk_y = input_y_i + 1j * input_y_q

    # 加上CMA预均衡收敛的信号长度
    qpsk_x = np.concatenate((qpsk_x[0:samples_per_symbol * num_update_cma], qpsk_x), axis=0)
    qpsk_y = np.concatenate((qpsk_y[0:samples_per_symbol * num_update_cma], qpsk_y), axis=0)

    # Initialization for CMA
    l_prior = floor((num_tap - 1) / 2)  # 计算抽头的左边界
    l_after = ceil((num_tap - 1) / 2)  # 计算抽头的右边界
    num_length = len(qpsk_x)

    num_output = floor((num_length - num_tap + 1) / samples_per_symbol)
    num_update = num_update_cma + num_output  # 计算全部的信号长度

    w_x = np.zeros((2 * num_tap, num_update + 1), dtype=np.complex128)
    w_y = np.zeros((2 * num_tap, num_update + 1), dtype=np.complex128)

    # 将各个FIR滤波器系数组合
    w_x[:, 0] = np.concatenate((initial_weight_xx, initial_weight_yx), axis=0).ravel()
    w_y[:, 0] = np.concatenate((initial_weight_xy, initial_weight_yy), axis=0).ravel()

    e_x = np.zeros(num_update, dtype=np.complex128)
    e_y = np.zeros(num_update, dtype=np.complex128)

    output_x = np.zeros((num_output, 1), dtype=np.complex128)
    output_y = np.zeros((num_output, 1), dtype=np.complex128)

    # 预均衡 Convergence
    # 采用CMA算法进行抽头系数预收敛

    ind_u = 0
    for ind in range(l_prior, l_prior + 1 + samples_per_symbol * (num_update_cma - 1), samples_per_symbol):
        prior = ind - l_prior - 1 if ind - l_prior - 1 >= 0 else None
        s = np.concatenate(
            (qpsk_x[ind + l_after:prior:-1], qpsk_y[ind + l_after:prior:-1]), axis=0)
        z_x = s.T @ w_x[:, ind_u]
        z_y = s.T @ w_y[:, ind_u]

        e_x[ind_u] = ref_power_cma - np.square(np.abs(z_x))  # 计算CMA的误差函数
        e_y[ind_u] = ref_power_cma - np.square(np.abs(z_y))

        w_x[:, ind_u + 1] = w_x[:, ind_u] + step_size_cma * z_x * np.conj(s) * e_x[ind_u]  # 更新CMA的抽头系数
        w_y[:, ind_u + 1] = w_y[:, ind_u] + step_size_cma * z_y * np.conj(s) * e_y[ind_u]

        ind_u = ind_u + 1

    # 设置不同星座图圆圈下的半径
    r1 = np.sqrt(2)
    r2 = np.sqrt(10)
    r3 = np.sqrt(18)
    r4 = np.sqrt(26)
    r5 = np.sqrt(34)
    r6 = np.sqrt(50)
    r7 = np.sqrt(58)
    r8 = np.sqrt(74)
    r9 = np.sqrt(98)

    # Realtime equalization
    # 采用RDE算法进行正式均衡
    ind_a = 0
    for ind in range(l_prior, l_prior + 1 + samples_per_symbol * (num_output - 1), samples_per_symbol):
        prior = ind - l_prior - 1 if ind - l_prior - 1 >= 0 else None
        s = np.concatenate((qpsk_x[ind + l_after:prior:-1], qpsk_y[ind + l_after:prior:-1]), axis=0)
        z_x = s.T @ w_x[:, ind_u]
        z_y = s.T @ w_y[:, ind_u]

        # 计算不同调制格式下的误差函数
        if bits_per_symbol == 2:  # QPSK
            # X偏振下的误差函数
            e_x[ind_u] = r1 - abs(z_x) ** 2
            # Y偏振下的误差函数
            e_y[ind_u] = r1 - abs(z_y) ** 2
        elif bits_per_symbol == 4:  # 16QAM
            # X
            if abs(z_x) ** 2 <= 0.5 * (r1 + r2):
                e_x[ind_u] = r1 - abs(z_x) ** 2
            elif abs(z_x) ** 2 <= 0.5 * (r2 + r3):
                e_x[ind_u] = r2 - abs(z_x) ** 2
            else:
                e_x[ind_u] = r3 - abs(z_x) ** 2
            # Y
            if abs(z_y) ** 2 <= 0.5 * (r1 + r2):
                e_y[ind_u] = r1 - abs(z_y) ** 2
            elif abs(z_y) ** 2 <= 0.5 * (r2 + r3):
                e_y[ind_u] = r2 - abs(z_y) ** 2
            else:
                e_y[ind_u] = r3 - abs(z_y) ** 2
        elif bits_per_symbol == 5:  # 32QAM
            # X
            if abs(z_x) ** 2 <= 0.5 * (r1 + r2):
                e_x[ind_u] = r1 - abs(z_x) ** 2
            elif abs(z_x) ** 2 <= 0.5 * (r2 + r3):
                e_x[ind_u] = r2 - abs(z_x) ** 2
            elif abs(z_x) ** 2 <= 0.5 * (r3 + r4):
                e_x[ind_u] = r3 - abs(z_x) ** 2
            elif abs(z_x) ** 2 <= 0.5 * (r4 + r5):
                e_x[ind_u] = r4 - abs(z_x) ** 2
            else:
                e_x[ind_u] = r5 - abs(z_x) ** 2
            # Y
            if abs(z_y) ** 2 <= 0.5 * (r1 + r2):
                e_y[ind_u] = r1 - abs(z_y) ** 2
            elif abs(z_y) ** 2 <= 0.5 * (r2 + r3):
                e_y[ind_u] = r2 - abs(z_y) ** 2
            elif abs(z_y) ** 2 <= 0.5 * (r3 + r4):
                e_y[ind_u] = r3 - abs(z_y) ** 2
            elif abs(z_y) ** 2 <= 0.5 * (r4 + r5):
                e_y[ind_u] = r4 - abs(z_y) ** 2
            else:
                e_y[ind_u] = r5 - abs(z_y) ** 2

        # 利用RDE算法对抽头系数进行更新
        w_x[:, ind_u + 1] = w_x[:, ind_u] + step_size_rde * e_x[ind_u] * z_x * np.conj(s)  # 更新CMA的抽头系数
        w_y[:, ind_u + 1] = w_y[:, ind_u] + step_size_rde * e_y[ind_u] * z_y * np.conj(s)
        ind_u = ind_u + 1

        output_x[ind_a] = z_x
        output_y[ind_a] = z_y
        ind_a = ind_a + 1

    # 祛除添加的预均衡信号，并去掉前后部分均衡效果不好的信号
    output_x = output_x[num_update_cma + 1999:-2000]
    output_y = output_y[num_update_cma + 1999:-2000]

    return output_x, output_y



import numpy as np
from numpy.fft import fftshift, fft, ifft, ifftshift


def optical_fiber_channel(tx_signal_x: np.ndarray,
                          tx_signal_y: np.ndarray,
                          sampling_rate: float,
                          span,
                          num_steps,
                          beta2,
                          delta_z,
                          gamma,
                          alpha,
                          L
                          ):
    """

    Args:
        tx_signal_x: 输入信号x
        tx_signal_y: 输入信号y
        sampling_rate:
        span: 一个跨度
        num_steps:
        beta2:
        delta_z: 单步步长
        gamma:

    Returns:

    """
    # step-1: Digital Backpropagation

    data_x = tx_signal_x
    data_y = tx_signal_y

    data_x_normal = data_x / np.sqrt(np.mean(abs(data_x) ** 2))
    data_y_normal = data_y / np.sqrt(np.mean(abs(data_y) ** 2))

    p1 = np.mean(abs(data_x_normal) ** 2)
    p2 = np.mean(abs(data_y_normal) ** 2)

    data_length = np.size(data_x)

    # 信号功率
    power_dbm_set = -6
    power_mw = 10 ** (power_dbm_set / 10)
    power_w = power_mw / 1000

    power_origin_x = np.mean(np.abs(data_x_normal) ** 2)
    power_origin_y = np.mean(np.abs(data_y_normal) ** 2)

    data_x = data_x_normal / np.sqrt(power_origin_x / power_w)
    data_y = data_y_normal / np.sqrt(power_origin_y / power_w)

    power_temp_x = np.mean(np.abs(data_x) ** 2)
    power_temp_y = np.mean(np.abs(data_y) ** 2)

    """ SSFM - Modeling """

    omega = np.arange(-sampling_rate / 2, sampling_rate / 2, sampling_rate / data_length)
    omega = np.reshape(omega, (-1, 1))

    data_fft_x = fftshift(fft(data_x, axis=0), axes=0) / np.size(data_x)  # 将数据变到频率并将零值移到原点
    data_fft_y = fftshift(fft(data_y, axis=0), axes=0) / np.size(data_y)

    for i in range(span):
        for j in range(num_steps):
            # 1/2 信号衰减
            data_fft_x = data_fft_x * np.exp(-alpha * (delta_z / 2))
            data_fft_y = data_fft_y * np.exp(-alpha * (delta_z / 2))

            # 1/2 光纤色散
            # X 偏振
            data_fft_x = data_fft_x * np.exp(-1j * (beta2 / 2) * (omega ** 2) * (delta_z / 2))
            # Y 偏振
            data_fft_y = data_fft_y * np.exp(-1j * (beta2 / 2) * (omega ** 2) * (delta_z / 2))

            # 光纤非线性
            # X 偏振
            data_t_x = ifft(ifftshift(data_fft_x, axes=0), axis=0) * np.size(data_fft_x)
            # Y 偏振
            data_t_y = ifft(ifftshift(data_fft_y, axes=0), axis=0) * np.size(data_fft_y)

            # X 偏振
            data_t_x = data_t_x * np.exp(8 / 9 * 1j * gamma * ((abs(data_t_y) ** 2) + (abs(data_t_x) ** 2)) * delta_z)
            # Y 偏振
            data_t_y = data_t_y * np.exp(8 / 9 * 1j * gamma * ((abs(data_t_y) ** 2) + (abs(data_t_x) ** 2)) * delta_z)

            data_fft_x = fftshift(fft(data_t_x, axis=0), axes=0) / np.size(data_t_x)
            data_fft_y = fftshift(fft(data_t_y, axis=0), axes=0) / np.size(data_t_y)

            # 1/2 光纤色散
            # X 偏振
            data_fft_x = data_fft_x * np.exp(-1j * (beta2 / 2) * (omega ** 2) * (delta_z / 2))
            # Y 偏振
            data_fft_y = data_fft_y * np.exp(-1j * (beta2 / 2) * (omega ** 2) * (delta_z / 2))

            # 1/2 信号衰减
            data_fft_x = data_fft_x * np.exp(-alpha * (delta_z / 2))
            data_fft_y = data_fft_y * np.exp(-alpha * (delta_z / 2))

        # == EDFA ==
        data_fft_x = data_fft_x * np.exp(alpha * L)
        data_fft_y = data_fft_y * np.exp(alpha * L)

    # step-2: Perturbation Theory
    # step-3: Neural Network

    tx_signal_x = ifft(ifftshift(data_fft_x, axes=0), axis=0) * np.size(data_fft_x)
    tx_signal_y = ifft(ifftshift(data_fft_y, axes=0), axis=0) * np.size(data_fft_y)

    return tx_signal_x, tx_signal_y, power_temp_x, power_temp_y



"""
Copyright (c) 2022 Beijing Jiaotong University
PhotLab is licensed under [Open Source License].
You can use this software according to the terms and conditions of the [Open Source License].
You may obtain a copy of [Open Source License] at: [https://open.source.license/]

THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.

See the [Open Source License] for more details.

Author: Chunyu Li
Created: 2022/9/6
Supported by: National Key Research and Development Program of China
"""

import numpy as np


def load_awgn(length_vector):
    """
    Generate AWGN for I and Q of two polarizations
    Args:
        length_vector: 输入的信号长度

    Returns:
        noise_x: 生成的X路信号的噪声
        noise_y: 生成的Y路信号的噪声
        noise_power: 噪声功率
    """
    seed = 11
    std_noise = 1  # 设置随机噪声方差为1

    # 生成的是高斯白噪声，分别对实部虚部进行高斯白噪声的生成

    rng = np.random.default_rng(seed + 41)  # 设置随机数种子
    noise_i_x = std_noise / 2 * rng.standard_normal((length_vector, 1))
    rng = np.random.default_rng(seed + 161)
    noise_q_x = std_noise / 2 * rng.standard_normal((length_vector, 1))
    rng = np.random.default_rng(seed + 217)
    noise_i_y = std_noise / 2 * rng.standard_normal((length_vector, 1))
    rng = np.random.default_rng(seed + 311)
    noise_q_y = std_noise / 2 * rng.standard_normal((length_vector, 1))

    # 合成噪声信号（实部+虚部）
    noise_x = noise_i_x + 1j * noise_q_x
    noise_y = noise_i_y + 1j * noise_q_y

    # noise_power：噪声功率计算噪声功率
    noise_power = np.mean(np.square(np.abs(noise_x)) + np.square(np.abs(noise_y)))

    return noise_x, noise_y, noise_power



"""
Copyright (c) 2022 Beijing Jiaotong University
PhotLab is licensed under [Open Source License].
You can use this software according to the terms and conditions of the [Open Source License].
You may obtain a copy of [Open Source License] at: [https://open.source.license/]

THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.

See the [Open Source License] for more details.

Author: Chunyu Li
Created: 2022/9/6
Supported by: National Key Research and Development Program of China
"""

import numpy as np
from scipy.signal import lfilter


def fine_synchronize(tr_data_x, re_training_x):
    # Method 1: the same to zhuge's
    # 采用自相关的同步算法，找到自相关性最大的位置

    # 错误信号= 取绝对值(filter滤波器(左右反转(取共轭(要同步的信号)),1,对应的正确的信号));
    matrix_error = np.abs(lfilter(np.fliplr(np.conj(re_training_x)).ravel(), 1, tr_data_x.ravel()))

    n = np.abs(np.sum(re_training_x * np.conj(re_training_x)))  # N的大小=取绝对值(求和(要同步的信号.*取共轭(要同步的信号)));
    matrix_error = matrix_error / n
    start_index = np.argmax(matrix_error)  # 找的最大的错误信号值，并确定最大错误值的位置序号
    start_index = start_index - len(re_training_x[0]) + 1  # 输出信号开始的同步序列号= 最大错误值的序号-要同步的信号的长度

    # % figure; plot(matric_error);

    return start_index



"""
Copyright (c) 2022 Beijing Jiaotong University
PhotLab is licensed under [Open Source License].
You can use this software according to the terms and conditions of the [Open Source License].
You may obtain a copy of [Open Source License] at: [https://open.source.license/]

THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.

See the [Open Source License] for more details.

Author: Chunyu Li
Created: 2022/9/6
Supported by: National Key Research and Development Program of China
"""

import numpy as np


def linewidth_induced_noise(num_symbols, over_sampling_rate, linewidth, symbol_rate):
    """

    Args:
        num_symbols: 输入信号长度
        over_sampling_rate: 采样率大小
        linewidth: 线宽
        symbol_rate: 符号速率

    Returns:
        phase_noise
    """
    # std_PhaseNoise= √(2πυ/采样率)   方差公式
    std_phase_noise = np.sqrt(2 * np.pi * linewidth / (symbol_rate * over_sampling_rate))

    # 生成的矩阵为均值为0，方差为std_phase_noise的数据
    rng = np.random.default_rng()  # Numpy Random Number Generator
    phase_noise_temp = std_phase_noise * rng.standard_normal((int(num_symbols * over_sampling_rate), 1))

    phase_noise = np.zeros((num_symbols, 1))  # 定一个num_symbols行，1列的0矩阵

    phase_noise[0] = phase_noise_temp[0]

    for idx in range(1, num_symbols):
        phase_noise[idx] = phase_noise[idx - 1] + phase_noise_temp[idx]

    #   对任意的t>s>=0，增量W(t)-W(s)~N(0,σ^2(t-s）），且s>0维纳过程的定义
    #  w =a+ b.*randn(m,n);
    # 其中：a为均值；
    #            b为标准差；需要开根号
    #            m为需要产生几行；
    #           n为需要产生几列.

    return np.reshape(phase_noise, (-1, 1))



"""
Copyright (c) 2022 Beijing Jiaotong University
PhotLab is licensed under [Open Source License].
You can use this software according to the terms and conditions of the [Open Source License].
You may obtain a copy of [Open Source License] at: [https://open.source.license/]

THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.

See the [Open Source License] for more details.

Author: Chunyu Li
Created: 2022/9/6
Supported by: National Key Research and Development Program of China
"""

from ..optical import load_awgn
import numpy as np


class GaussianNoise:
    def __init__(self, osnr, sampling_rate):
        self.osnr = osnr
        self.sampling_rate = sampling_rate

    def add(self, signal_x, signal_y):
        # 生成均值为0，方差为1的随机噪声,此处直接产生两个偏振的噪声
        noise_x, noise_y, noise_power = load_awgn(len(signal_x))

        # 计算当前信号功率
        original_avg_power = np.mean(np.square(np.abs(signal_x)) + np.square(np.abs(signal_x)))

        osnr = np.power(10, self.osnr / 10)  # 将OSNR的单位由dB转为常量单位

        # 先计算OSNR对应的SNR，再通过SNR计算所需要达到的目标的信号功率，12.5e9为信号的中心频率，此处是根据公式计算，可参考通信原理或者百度或者论文
        target_avg_power = noise_power * osnr * 12.5e9 / self.sampling_rate

        # 改变当前信号功率为目标功率
        signal_x = np.sqrt(target_avg_power / original_avg_power) * signal_x
        signal_y = np.sqrt(target_avg_power / original_avg_power) * signal_y

        # 对信号添加随机噪声
        signal_x = noise_x + signal_x
        signal_y = noise_y + signal_y

        return signal_x, signal_y



"""
Copyright (c) 2022 Beijing Jiaotong University
PhotLab is licensed under [Open Source License].
You can use this software according to the terms and conditions of the [Open Source License].
You may obtain a copy of [Open Source License] at: [https://open.source.license/]

THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.

See the [Open Source License] for more details.

Author: Chunyu Li
Created: 2022/9/6
Supported by: National Key Research and Development Program of China
"""

import numpy as np
from scipy.special import erfcinv


def ber_count(received_sequence, transmit_sequence):
    num_transmission = len(transmit_sequence)  # 计算信号长度
    num_receive = len(received_sequence)

    # 调整信号为1列
    received_sequence = np.reshape(received_sequence, (-1, 1))
    transmit_sequence = np.reshape(transmit_sequence, (-1, 1))

    # 计算发射端信号和接收端信号的长度差，标记为需要丢掉的长度
    num_abandon = num_transmission - num_receive

    total_error_sequence = np.zeros((num_abandon + 1, 1))

    for ind in range(num_abandon + 1):  # 这个循环是在计算每种情况下的错误码元长度
        temp = received_sequence[0:num_receive] - transmit_sequence[ind:num_receive + ind]
        total_error_sequence[ind] = len(np.argwhere(temp != 0))

    total_error = np.min(total_error_sequence)  # 定义错误码元数最少的为信号的错误码元数

    total_bits = num_receive
    ber = total_error / total_bits  # 用错误码元数除以全部的符号数，就是误码率
    q = np.sqrt(2) * erfcinv(2 * ber)  # 利用公式利用BER计算出Q因子
    q_db = 20 * np.log10(q)  # 调整单位为dB

    return ber, q_db


def ber_estimate(tx_scm_matrix, rx_scm_matrix, bits_per_symbol):
    # Count the number of error symbols.
    length_symbols = np.size(tx_scm_matrix)
    error_sym = 0
    delta_tr_matrix = rx_scm_matrix - tx_scm_matrix
    delta_iq = np.concatenate((np.abs(np.real(delta_tr_matrix)), np.abs(np.imag(delta_tr_matrix))), axis=1)
    for sym_num in range(length_symbols):
        if delta_iq[sym_num][0] > 1:
            error_sym = error_sym + 1
        if delta_iq[sym_num][1] > 1:
            error_sym = error_sym + 1
    # Estimate the BER value
    ber = error_sym / (2 * length_symbols) / (bits_per_symbol / 2)
    q = np.sqrt(2) * erfcinv(2 * ber)  # 利用公式利用BER计算出Q因子
    q_db = 20 * np.log10(q)  # 调整单位为dB

    return ber, q_db



import numpy as np
from numpy.fft import fftshift, fft, ifft, ifftshift


def cdc(re_x, re_y, power_temp_x, power_temp_y, sampling_rate, beta2, span, L):
    power_temp2_x = np.mean(abs(re_x) ** 2)
    power_temp2_y = np.mean(abs(re_y) ** 2)

    re_x = re_x / (np.sqrt(power_temp2_x / power_temp_x))
    re_y = re_y / (np.sqrt(power_temp2_y / power_temp_y))

    data_length = np.size(re_x)
    omega = np.arange(-sampling_rate / 2, sampling_rate / 2, sampling_rate / data_length)
    omega = np.reshape(omega, (-1, 1))

    re_x_fre = fftshift(fft(re_x)) / np.size(re_x)
    re_y_fre = fftshift(fft(re_y)) / np.size(re_y)

    re_x_fre = re_x_fre * np.exp(1j * (beta2 / 2) * (omega ** 2) * (span * L))
    re_y_fre = re_y_fre * np.exp(1j * (beta2 / 2) * (omega ** 2) * (span * L))

    re_x = ifft(ifftshift(re_x_fre)) * np.size(re_x)
    re_y = ifft(ifftshift(re_y_fre)) * np.size(re_y)

    return re_x, re_y



"""
Copyright (c) 2022 Beijing Jiaotong University
PhotLab is licensed under [Open Source License].
You can use this software according to the terms and conditions of the [Open Source License].
You may obtain a copy of [Open Source License] at: [https://open.source.license/]

THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.

See the [Open Source License] for more details.

Author: Chunyu Li
Created: 2022/9/6
Supported by: National Key Research and Development Program of China
"""

from .pulse_shaping import PulseShaper
from .dac_noise import DacNoise
from .phase_noise import PhaseNoise
from .gaussian_noise import GaussianNoise
from .frame_syncer import FrameSyncer
from .adaptive_equalizer import AdaptiveEqualizer
from .bps import BPS
from .ber import BER
from .modem import Modem
from .rx_simple import add_freq_offset, add_iq_imbalance, add_adc_noise, iq_freq_offset_and_compensation
